/*
    tetris is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    tetris is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with tetris.  If not, see <http://www.gnu.org/licenses/>.

   ------------------------------------------------------------------------------->

  Author: Federico Salotti
  Organization/Company: Mr_Robot_Computers_Repair
  Country: Italy Barga(LU)
  release: 1.0.0 (05 November 2020 22:55)

*/


#include <SFML/Graphics.hpp>
#include <time.h>
#include <fstream>
#include <cstring>
#include "symbols.h"
#include "Cell.h"

float fall_interval=0.6f;
float start_fall_interval=fall_interval;
int grid[GRID_H][GRID_W]={0};
int next_grid[4][4]={0};
sf::Clock tick;
sf::Clock level_tick;
bool moveL=false;
bool moveR=false;
bool rotation=false;
bool pause=false;
bool game_over=false;
unsigned lv = 1;
std::string bg_string = "bg/bg1.png";
unsigned long score=0;
unsigned long hscore=0;

int tetromino[8][4][4]=
{
	// PLACEHOLDER //
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,

	// I //
	0,1,0,0,
	0,1,0,0,
	0,1,0,0,
	0,1,0,0,

	// Z - LEFT//
	1,0,0,0,
	1,1,0,0,
	0,1,0,0,
	0,0,0,0,

	// Z - RIGHT
	0,1,0,0,
	1,1,0,0,
	1,0,0,0,
	0,0,0,0,

	// [] //
	1,1,0,0,
	1,1,0,0,
	0,0,0,0,
	0,0,0,0,

	// |- //
	1,0,0,0,
	1,1,0,0,
	1,0,0,0,
	0,0,0,0,

	// L //
	1,0,0,0,
	1,0,0,0,
	1,1,0,0,
	0,0,0,0,

	// L - LEFT //

	0,1,0,0,
	0,1,0,0,
	1,1,0,0,
	0,0,0,0

};

sf::Color bg1_palette[] =
{
	sf::Color::Transparent,
	sf::Color(0xff,0x71,0xce),
	sf::Color(0x01,0xcd,0xfe),
	sf::Color(0x5f,0xfa,0xa1),
	sf::Color(0xb9,0x67,0xff),
	sf::Color(0xff,0x9b,0x83),
	sf::Color(0xff,0xfb,0x96),
	sf::Color(0xff,0xda,0xf5)
};

sf::Color bg2_palette[] =
{
	sf::Color::Transparent,
	sf::Color(0x00,0x00,0x3b),
	sf::Color(0x33,0x00,0x3b),
	sf::Color(0x66,0x00,0x3b),
	sf::Color(0x99,0x00,0x3b),
	sf::Color(0xcc,0x00,0x3b),
	sf::Color(0xff,0x00,0x3b),
	sf::Color(0x03,0x5e,0xe8)
};

sf::Color bg3_palette[] =
{
	sf::Color::Transparent,
	sf::Color(0xd0,0xee,0x88),
	sf::Color(0xe7,0x9a,0x9a),
	sf::Color(0xf2,0xea,0xea),
	sf::Color(0xb0,0xef,0xd1),
	sf::Color(0xfa,0xb8,0xca),
	sf::Color(0xfb,0xcd,0xed),
	sf::Color(0xbd,0xbd,0xbd)
};

sf::Color bg4_palette[] =
{
	sf::Color::Transparent,
	sf::Color(0x6e,0x17,0x61),
	sf::Color(0x90,0x24,0xac),
	sf::Color(0xc2,0x29,0x41),
	sf::Color(0xdd,0x5c,0x9d),
	sf::Color(0xe6,0xad,0x6c),
	sf::Color::Yellow,
	sf::Color::Green
};

sf::Color bg5_palette[] =
{
	sf::Color::Transparent,
	sf::Color(0xff,0xa3,0x22),
	sf::Color(0xbb,0x14,0x14),
	sf::Color(0x04,0x21,0x34),
	sf::Color(0x15,0x26,0x10),
	sf::Color::Black,
	sf::Color(0x4f,0x01,0x73),
	sf::Color(0x21,0x67,0x44)
};


sf::Color colors[]
{
	sf::Color::Transparent,
	sf::Color(0xff,0x71,0xce),
	sf::Color(0x01,0xcd,0xfe),
	sf::Color(0x5f,0xfa,0xa1),
	sf::Color(0xb9,0x67,0xff),
	sf::Color(0xff,0x9b,0x83),
	sf::Color(0xff,0xfb,0x96),
	sf::Color(0xff,0xda,0xf5)
};

int type=0;
int current_x;
int current_y;

int next_type=0;
int start_x=GRID_W/2;
int start_y=0;

int main()
{
	sf::Texture lv_texture;
	lv_texture.loadFromFile(bg_string);

	sf::Sprite bg;
	bg.setTexture(lv_texture);

	sf::Texture pink_texture;
	pink_texture.loadFromFile("textures/purple.png");

	sf::Font font;
	font.loadFromFile("fonts/Cybrpnuk2.ttf");

	sf::Text game_over_ui("Game Over",font);
	game_over_ui.setCharacterSize(60);
	game_over_ui.setPosition((CELL_SIZE*GRID_W)/2,(CELL_SIZE*GRID_W)/2);

	sf::Text score_ui("SCORE  "+std::to_string(score),font);
	score_ui.setPosition(CELL_SIZE*GRID_W+50,CELL_SIZE*GRID_W-100);

	sf::Text lv_ui(" LV  "+std::to_string(lv),font);
	lv_ui.setPosition(CELL_SIZE*GRID_W+50,CELL_SIZE*GRID_W-160);

	sf::RenderWindow window(sf::VideoMode(CELL_SIZE*GRID_W+300,CELL_SIZE*GRID_H),"TETRIS");
	srand(time(nullptr));
	Cell cell;

	auto get_hscore = [&]()
	{
		std::ifstream reader("high_score");
		std::string hscore_str;
		std::getline(reader,hscore_str);
		hscore = std::stoi(hscore_str);
		reader.close();
	}; get_hscore();

	auto set_hscore = [&]()
	{
		std::ofstream writer;
		writer.open("high_score");
		writer << score;
		writer.close();
	};

	sf::Text high_score_ui("HIGH SCORE \n\t"+std::to_string(hscore),font);
	high_score_ui.setPosition(CELL_SIZE*GRID_W+50,CELL_SIZE*GRID_W-40);

	auto set_palette = [&]()
	{
		if(bg_string.compare("bg/bg1.png") == 0)
		{
			memcpy(&colors,&bg1_palette,sizeof(bg1_palette));
			score_ui.setFillColor(sf::Color(0x97,0x00,0xcc));
			high_score_ui.setFillColor(sf::Color(0x97,0x00,0xcc));
			lv_ui.setFillColor(sf::Color(0x97,0x00,0xcc));
		}

		if(bg_string.compare("bg/bg2.png") == 0)
		{
			memcpy(&colors,&bg2_palette,sizeof(bg2_palette));
			score_ui.setFillColor(sf::Color(0x2a,0x4e,0x2a));
			high_score_ui.setFillColor(sf::Color(0x2a,0x4e,0x2a));
			lv_ui.setFillColor(sf::Color(0x2a,0x4e,0x2a));
		}

		if(bg_string.compare("bg/bg3.png") == 0)
		{
			memcpy(&colors,&bg3_palette,sizeof(bg3_palette));
			score_ui.setFillColor(sf::Color(0x45,0x5c,0x72));
			high_score_ui.setFillColor(sf::Color(0x45,0x5c,0x72));
			lv_ui.setFillColor(sf::Color(0x45,0x5c,0x72));
		}

		if(bg_string.compare("bg/bg4.png") == 0)
		{
			memcpy(&colors,&bg4_palette,sizeof(bg4_palette));
			score_ui.setFillColor(sf::Color(0x76,0x6a,0x80));
			high_score_ui.setFillColor(sf::Color(0x76,0x6a,0x80));
			lv_ui.setFillColor(sf::Color(0x76,0x6a,0x80));
		}

		if(bg_string.compare("bg/bg5.png") == 0)
		{
			memcpy(&colors,&bg5_palette,sizeof(bg5_palette));
			score_ui.setFillColor(sf::Color(0x89,0x6d,0x43));
			high_score_ui.setFillColor(sf::Color(0x89,0x6d,0x43));
			lv_ui.setFillColor(sf::Color(0x89,0x6d,0x43));
		}

	};set_palette();

	auto set_lv_bg = [&]()
	{
		if(lv == 1)
		{
			bg_string = "bg/bg1.png";
			cell.setGridColor(sf::Color(0x97,0x00,0xcc));
		}
		if(lv == 2)
		{	bg_string = "bg/bg2.png";
			cell.setGridColor(sf::Color(0x2a,0x4e,0x2a));
		}

		if(lv == 3)
		{
			bg_string = "bg/bg3.png";
			cell.setGridColor(sf::Color(0x45,0x5c,0x72));
		}
		if(lv == 4)
		{
			bg_string = "bg/bg4.png";
			cell.setGridColor(sf::Color(0x76,0x6a,0x80));
		}
		if(lv >= 5)
		{
			bg_string = "bg/bg5.png";
			cell.setGridColor(sf::Color(0x89,0x6d,0x43));
		}

		lv_texture.loadFromFile(bg_string);
		bg.setTexture(lv_texture);
		set_palette();
	};set_lv_bg();


	auto drawGrid = [&]()
	{
		for(int y=0; y < GRID_H; ++y)
		{
			for(int x=0; x < GRID_W; ++x)
			{
				cell.setPosition(x*CELL_SIZE,y*CELL_SIZE);
				if(grid[y][x] == 0)cell.setColor(sf::Color::Transparent);
				else cell.setColor(colors[grid[y][x]]);
				cell.draw(window);
			}
		}
	};

	auto drawNextGrid = [&]()
	{
		for(int y=0; y < 4; ++y)
		{
			for(int x=0; x < 4; ++x)
			{
				cell.setPosition(x*CELL_SIZE+500,y*CELL_SIZE+50);
				if(next_grid[y][x] == 0)cell.setColor(sf::Color::Transparent);
				else cell.setColor(colors[next_grid[y][x]]);
				cell.draw(window);
			}
		}
	};


	auto generateTetromino = [&]()
	{
		if(next_type != 0) type=next_type;
		else type=rand()%7 + 1;
		next_type=rand()%7 + 1;
		current_x=GRID_W / 2;
		current_y=0;

	};generateTetromino();

	auto drawTetromino = [&]()
	{
//		cell.setColor(colors[type]);
//		cell.setTexture(pink_texture);
		for(int y=0; y < 4; ++y)
		{
			for(int x=0; x < 4; ++x)
			{
				if(tetromino[type][y][x])
				{
					cell.setPosition((current_x+x)*CELL_SIZE,(current_y+y)*CELL_SIZE);
					cell.draw(window);
				}
			}
		}

	};

	auto drawNextTetromino = [&]()
	{
		fall_interval = start_fall_interval;

		for(int y=0; y < 4; ++y)
			for(int x=0; x < 4; ++x)
				next_grid[y][x]=0;

		cell.setColor(colors[type]);

		for(int y=0; y < 4; ++y)
			for(int x=0; x < 4; ++x)
				if(tetromino[next_type][y][x])
				{
					if(next_type == 0)
						next_grid[y][x]=next_type + 1;
					else
						next_grid[y][x]=next_type;
				}
	};

	auto check_collision = [&]()
	{
		for(int y=0; y < 4; ++y)
		{
			for(int x=0; x < 4; ++x)
			{
				if(!tetromino[type][y][x]) continue;
				if(grid[current_y+y][current_x+x]) return true;
				if(current_x + x < 0) return true;
				if(current_x + x >= GRID_W) return true;
				if(current_y + y >= GRID_H) return true;
			}
		}

		return false;
	};

	auto line_check = [&]()
	{
		int start=GRID_H-1;

		for(int end_grid=GRID_H-1; end_grid >= 0; --end_grid)
		{
			int line_counter=0;

			for(int x=0; x < GRID_W; ++x)
				if(grid[end_grid][x]) line_counter++;

			if(line_counter < GRID_W) // se la riga non è completa viene mantenuta
						  // quindi copiata
			{
				for(int x=0; x < GRID_W; ++x)
					grid[start][x]=grid[end_grid][x];
				start--;
			}
			else
			{
				score+=75;
				// Skippa la riga eliminandola
			}
		}

	};

	auto check_game_over = [&]()
	{
		for(int x=0; x < GRID_W; ++x)
			if(grid[0][x])
				game_over=true;
	};

	auto reset_grid = [&]()
	{
		for(int y=0; y < GRID_H; ++y)
			for(int x=0; x < GRID_W; ++x)
				grid[y][x]=0;
	};

	auto fall = [&]()
	{
		current_y++;
		if(check_collision())
		{
			current_y--;
			for(int y=0; y < 4; ++y)
			{
				for(int x=0; x < 4; ++x)
				{
					if(tetromino[type][y][x])
					{
						if(type == 0)
							grid[current_y+y][current_x+x]=type + 1;
						else
							grid[current_y+y][current_x+x]=type;
					}
				}
			}
		line_check();
		generateTetromino();
		}
	};

	auto rotate = [&]()
	{
		int tetromino_length=0;

		for(int y=0; y < 4; ++y)
		{
			for(int x=0; x < 4; ++x)
			{
				if(tetromino[type][y][x])
					tetromino_length=std::max(std::max(x,y)+1,tetromino_length);

			}
		}

		int tetromino_matrix[4][4]={0};

		for(int y=0; y < tetromino_length; ++y)
			for(int x=0; x < tetromino_length; ++x)
				if(tetromino[type][y][x]) tetromino_matrix[tetromino_length - 1 - x][y]=1;


		for(int y=0; y < 4; ++y)
			for(int x=0; x < 4; ++x)
				tetromino[type][y][x] = tetromino_matrix[y][x];

		rotation=false;
	};


	while(window.isOpen())
	{
		if(level_tick.getElapsedTime().asSeconds() > LV_CHANGE_INTERVAL)
		{
			lv += 1;
			set_lv_bg();
			set_palette();
			fall_interval = start_fall_interval;
			if(fall_interval >= 0.2f) fall_interval -= 0.1f;
			start_fall_interval = fall_interval;
			level_tick.restart();
		}

		lv_ui.setString("LV  "+std::to_string(lv));
		score_ui.setString("SCORE  "+std::to_string(score));
		if(score > hscore) high_score_ui.setString("HIGH SCORE\n\t"+std::to_string(score));
		else high_score_ui.setString("HIGH SCORE\n\t"+std::to_string(hscore));

		sf::Event event;

		while(window.pollEvent(event))
		{
			if(event.type == sf::Event::Closed) window.close();

			if(sf::Keyboard::isKeyPressed(sf::Keyboard::W)) rotation=true;

			else if(sf::Keyboard::isKeyPressed(sf::Keyboard::A))
			{
				moveL=true;
			}
			else if(sf::Keyboard::isKeyPressed(sf::Keyboard::D))
			{
				moveR=true;
			}

			else if(sf::Keyboard::isKeyPressed(sf::Keyboard::S))
			{
				if(fall_interval/2 == start_fall_interval/2) fall_interval/=2;
			}

			else if(event.type == sf::Event::KeyPressed)
			{
				if(event.key.code == sf::Keyboard::Space)
				{
					if(!game_over) pause=!pause;
					else
					{
						reset_grid();
						if(score > hscore) set_hscore();
						score = 0;
						get_hscore();
						lv = 1;
						bg_string = "bg/bg1.png";
						fall_interval = ORIGINAL_INTERVAL;
						start_fall_interval = ORIGINAL_INTERVAL;
						set_lv_bg();
						game_over=false;
					}
				}
			}

			else if(sf::Event::KeyReleased)
			{
				if(event.key.code == sf::Keyboard::S) fall_interval=start_fall_interval;
			}
		}

		if(moveL)
		{
			if(!pause)
			{
				current_x--;
				if(check_collision()) current_x++;
				moveL=false;
			}
		}

		if(moveR)
		{
			if(!pause)
			{
				current_x++;
				if(check_collision()) current_x--;
				moveR=false;
			}
		}

		if(rotation)
		{
			if(!pause)
			{
				rotate();
				if(check_collision()) {rotate(),rotate();rotate();}
			}
		}

		if(tick.getElapsedTime().asSeconds() > fall_interval)
		{
			if(!pause) fall();
			tick.restart();
		}

		window.clear();
		window.draw(bg);
		if(!game_over)
		{
			drawGrid();
			check_game_over();
			drawNextGrid();
			drawNextTetromino();
			drawTetromino();
			window.draw(score_ui);
			window.draw(high_score_ui);
			window.draw(lv_ui);

		}
		else
		{
			window.draw(game_over_ui);
			lv = 1;
			level_tick.restart();
		}
		window.display();
	}
	return 0;
}
